/*  read2b.c            eej    2/97  */
/*                                   */
/*  reads a PDATS2b trace and        */
/*  writes bytes and dinero to stdout */

#include <stdio.h>

#define DATA_STREAMS 2

unsigned long DataAddr[DATA_STREAMS];
unsigned long PC;

int DefaultInstrStride = 4;
int DefaultInstrShift = 2;  /* log2 of DefaultInstrStride */
int DEBUG = 0;


/*********************************************************************/

long
NegNum(unsigned long num, int size)
{
  switch (size) {
  case 0:
  case 4:
  case 5:
  case 6:
  case 7:
    return 0;
  case 1: return (num & 0x80);
  case 2: return (num & 0x8000);
  case 3: return (num & 0x80000000);
  default:
    fprintf(stderr, "\nread2b:  illegal size (%d) passed to NegNum\n",
	    size);
    exit(-1);
  } 
}

/****************************************************************/

#define neg1 0xffffff00
#define neg2 0xffff0000

SignExtend(num,size)
long num;
char size;
{
    if(size==1) return (num | neg1);
    if(size==2) return (num | neg2);
    else return num;
}

/**************************************************************/

PD2b2DIN()
{
  int i, c, d, addrCode, NumRep, addr, dataStream;
  long addrOffset;

  while ((c = getchar()) != EOF) {
    printf("%02x ", c);
    if (c & 0x80) {  /* data record */
      NumRep = (c & 0x30) >> 4;
      dataStream = (c & 8) >> 3;
      addr = DataAddr[dataStream];

      addrCode = (c & 0x07);

      /* now compute offset, update and output address */
      switch (addrCode) {
      case 0:  addr += 4; break;
      case 4:  addr -= 8; break;
      case 5:  addr -= 4; break;
      case 6:  break;
      case 7:  addr += 8; break;
      case 1:
      case 2: 
	addrOffset = 0;
	for(i=addrCode; i>0; i--) {
	  d = getchar();
	  printf("%02x", d);
	  addrOffset |= d << ((i-1) << 3);
	}
	if (NegNum(addrOffset,addrCode))
	  addrOffset= SignExtend(addrOffset,addrCode);
	addr += addrOffset;
	break;
      case 3:
	fread((char *) &addrOffset, 4, 1, stdin);
	printf("%08x", addrOffset);
	if (NegNum(addrOffset,addrCode))
	  addrOffset = SignExtend(addrOffset,addrCode);
	addr += addrOffset;
	break;
      }
      printf("\n");
      DataAddr[dataStream] = addr;
      if (c & 0x40) printf("\t1 ");  /* write */
      else printf("\t0 ");
      printf("%x\n", addr);

      /* now emit sequential I-fetches if any */
      for (i=0; i<NumRep; i++) 
	printf("\t2 %x\n", PC += DefaultInstrStride);
    }
    else {  /* instr record */
      addrCode = (c & 0x03);
      if (addrCode == 0) {
	addrOffset = DefaultInstrStride;
	printf("\n");
      }
      else {
	if (addrCode == 1 || addrCode == 2) {
	  addrOffset = 0;
	  for(i=addrCode; i>0; i--) {
	    d = getchar();
	    printf("%02x", d);
	    addrOffset |= d << ((i-1) << 3);
	  }
	  printf("\n");
	}
	else {
	  fread((char *) &addrOffset, 4, 1, stdin);
	  printf("%08x\n", addrOffset);
	}
	if (NegNum(addrOffset, addrCode))
	  addrOffset= SignExtend(addrOffset, addrCode);
	addrOffset <<= DefaultInstrShift;
      }

      printf("\t2 %x\n", PC += addrOffset);

      NumRep = (c >> 2) & 31;
      for (i=0; i<NumRep; i++) 
	printf("\t2 %x\n", PC += DefaultInstrStride);
    }
  }
}

/********************************************************************/

initial()
{
  int i, j;
  
  DataAddr[0] = 
  DataAddr[1] = 0;
  
  PC = 0;
}

/********************************************************************/

main()
{
    int c;

    if (getchar() != 0xff) {
	fprintf(stderr, "read2b: file not in PDATS format\n");
	exit(-1);
    }
    if (getchar() != 0x8a) {
	fprintf(stderr, "read2b: file not in PDATS 2 format\n");
	exit(-2);
    }
    initial();
    PD2b2DIN();
}

